/*
 * Copyright 2013 Mariusz Kopylec
 *
 * This file is part of SQL Mapper.
 *
 * SQL Mapper is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * SQL Mapper is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Foobar. If not, see http://www.gnu.org/licenses/
 */
package net.sourceforge.kopylec.sqlmapper.core;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * A basic row mapper which maps results of an SQL query to a list of objects
 * (entities). This class should be extended in order to process a result of an
 * SQL query which returns <b>more than one entity</b>.
 *
 * @author Mariusz Kopylec
 * @param <T> a type to map an SQL query result to
 * @version 1.0
 */
public abstract class MultipleResultsMapper<T> extends ResultMapper<List<T>> {

    private List<T> results;

    @Override
    List<T> getQueryResult() throws SQLException {
        results = new ArrayList<>();
        while (getResultSet().next()) {
            mapRow();
        }
        return results;
    }

    /**
     * Adds an entity to results list if it is not already added. For the method
     * to work correctly an entity must override {@code equals} and
     * {@code hashCode} methods.
     *
     * @param entity entity to merge with results list
     * @return an entity added to results list
     */
    protected final T mergeWithResults(T entity) {
        int entityIndex = results.indexOf(entity);
        if (entityIndex > -1) {
            return results.get(entityIndex);
        } else {
            results.add(entity);
            return entity;
        }
    }

    /**
     * Adds an entity to results list.
     *
     * @param entity entity to add
     */
    protected final void addToResults(T entity) {
        results.add(entity);
    }

    /**
     * Gets the current results list. The results list contains already mapped
     * SQL query results.
     *
     * @return results list
     */
    protected final List<T> getResults() {
        return results;
    }

    /**
     * Maps a current row from result set to the results list.
     *
     * @throws SQLException if a database access error occurs
     */
    protected abstract void mapRow() throws SQLException;
}
