/*
 * Copyright 2013 Mariusz Kopylec
 *
 * This file is part of SQL Mapper.
 *
 * SQL Mapper is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * SQL Mapper is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Foobar. If not, see http://www.gnu.org/licenses/
 */
package net.sourceforge.kopylec.sqlmapper.core;

import java.math.BigDecimal;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Date;
import java.util.List;
import net.sourceforge.kopylec.sqlmapper.exception.QueryResultException;

/**
 * An abstract class for {@link SingleResultMapper} and
 * {@link MultipleResultsMapper}
 *
 * @author Mariusz Kopylec
 * @param <T> a type to map an SQL query result to
 * @version 1.0
 */
abstract class ResultMapper<T> {

    private ResultSet resultSet;

    abstract T getQueryResult() throws SQLException, QueryResultException;

    /**
     * Adds an entity to entities list if it is not already added. For the
     * method to work correctly an entity must override {@code equals} and
     * {@code hashCode} methods.
     *
     * @param <E> an entity type
     * @param entity entity to merge with entities list
     * @param entities list of entities
     * @return an entity added to entities list
     */
    protected final <E> E merge(E entity, List<E> entities) {
        int entityIndex = entities.indexOf(entity);
        if (entityIndex > -1) {
            return entities.get(entityIndex);
        } else {
            entities.add(entity);
            return entity;
        }
    }

    /**
     * Gets the {@code java.sql.ResultSet} with a cursor set to row that is
     * currently being processed.
     *
     * @return a result set created from SQL query
     */
    protected final ResultSet getResultSet() {
        return resultSet;
    }

    void setResultSet(ResultSet resultSet) {
        this.resultSet = resultSet;
    }

    /**
     * Checks if the specified column value from the result set is null.
     *
     * @param column result set column name
     * @return {@code true} if value is null, otherwise {@code false}
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final boolean isNull(String column) throws SQLException {
        return getObject(column) == null;
    }

    /**
     * Checks if the specified column value from the result set is not null.
     *
     * @param column result set column name
     * @return {@code true} if value is not null, otherwise {@code false}
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final boolean isNotNull(String column) throws SQLException {
        return !isNull(column);
    }

    /**
     * Gets an integer value of the specified column from the result set.
     *
     * @param column result set column name
     * @return an integer value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Integer getInteger(String column) throws SQLException {
        return getObject(column) == null ? null : resultSet.getInt(column);
    }

    /**
     * Gets a string value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a string value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final String getString(String column) throws SQLException {
        return resultSet.getString(column);
    }

    /**
     * Gets an object value of the specified column from the result set.
     *
     * @param column result set column name
     * @return an object value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Object getObject(String column) throws SQLException {
        return resultSet.getObject(column);
    }

    /**
     * Gets a date value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a date value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Date getDate(String column) throws SQLException {
        return resultSet.getDate(column);
    }

    /**
     * Gets a double value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a double value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Double getDouble(String column) throws SQLException {
        return getObject(column) == null ? null : resultSet.getDouble(column);
    }

    /**
     * Gets a boolean value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a boolean value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Boolean getBoolean(String column) throws SQLException {
        return getObject(column) == null ? null : resultSet.getBoolean(column);
    }

    /**
     * Gets a big decimal value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a big decimal value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final BigDecimal getBigDecimal(String column) throws SQLException {
        return resultSet.getBigDecimal(column);
    }

    /**
     * Gets a long value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a long value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Long getLong(String column) throws SQLException {
        return getObject(column) == null ? null : resultSet.getLong(column);
    }

    /**
     * Gets a timestamp value of the specified column from the result set.
     *
     * @param column result set column name
     * @return a timestamp value
     * @throws SQLException if the column name is not valid; if a database
     * access error occurs or this method is called on a closed result set
     */
    protected final Timestamp getTimestamp(String column) throws SQLException {
        return resultSet.getTimestamp(column);
    }
}
