/*
 * Copyright 2013 Mariusz Kopylec
 *
 * This file is part of SQL Mapper.
 *
 * SQL Mapper is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * SQL Mapper is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Foobar. If not, see http://www.gnu.org/licenses/
 */
package net.sourceforge.kopylec.sqlmapper.core;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import javax.sql.DataSource;
import net.sourceforge.kopylec.sqlmapper.exception.QueryResultException;
import net.sourceforge.kopylec.sqlmapper.exception.SQLMapperException;

/**
 * A main class of {@code sql-mapper}. SQL runner contains methods to run SQL
 * queries and updates.
 *
 * @author Mariusz Kopylec
 * @version 1.0
 */
public class SQLRunner {

    private final DataSource dataSource;
    private Connection connection;
    private boolean transactionActive;
    private boolean transactionFailed;

    public SQLRunner(DataSource dataSource) {
        this.dataSource = dataSource;
        transactionActive = false;
        transactionFailed = false;
    }

    /**
     * Runs an SQL query and maps its result to a <b>single object</b>.
     *
     * @param <E> result object type
     * @param query SQL select statement
     * @param resultClass result object class
     * @param parameters query parameters
     * @return SQL query result mapped to a single object
     * @throws SQLMapperException if a database access error occurs; if query
     * returns more than one column; if query returns more than one result
     */
    public <E> E selectOne(String query, Class<E> resultClass, Object... parameters) throws SQLMapperException {
        List<E> results = select(query, resultClass, true, parameters);
        return results.isEmpty() ? null : results.get(0);
    }

    /**
     * Runs an SQL query and maps its results to a <b>single entity</b>.
     *
     * @param <E> result entity type
     * @param query SQL select statement
     * @param mapper mapper which processes SQL query results
     * @param parameters query parameters
     * @return SQL query results mapped to a single entity
     * @throws SQLMapperException if a database access error occurs; if query
     * returns more than one entity
     */
    @SuppressWarnings("unchecked")
    public <E> E selectOne(String query, SingleResultMapper<E> mapper, Object... parameters) throws SQLMapperException {
        return (E) select(query, mapper, parameters);
    }

    /**
     * Runs an SQL query and maps its results to a <b>list of objects</b>.
     *
     * @param <E> result object type
     * @param query SQL select statement
     * @param resultClass result object class
     * @param parameters query parameters
     * @return SQL query results mapped to a list of objects
     * @throws SQLMapperException if a database access error occurs; if query
     * returns more than one column
     */
    public <E> List<E> selectMany(String query, Class<E> resultClass, Object... parameters) throws SQLMapperException {
        return select(query, resultClass, false, parameters);
    }

    /**
     * Runs an SQL query and maps its results to a <b>list of entities</b>.
     *
     * @param <E> result entity type
     * @param query SQL select statement
     * @param mapper mapper which processes SQL query results
     * @param parameters query parameters
     * @return SQL query results mapped to a list of entities
     * @throws SQLMapperException if a database access error occurs
     */
    @SuppressWarnings("unchecked")
    public <E> List<E> selectMany(String query, MultipleResultsMapper<E> mapper, Object... parameters) throws SQLMapperException {
        return (List<E>) select(query, mapper, parameters);
    }

    /**
     * Runs an SQL insert, update or delete statement.
     *
     * @param update SQL statement
     * @param parameters update parameters
     * @throws SQLMapperException if a database access error occurs
     */
    public void update(String update, Object... parameters) throws SQLMapperException {
        PreparedStatement ps = null;
        ResultSet rs = null;
        try {
            failIfConnectionIsClosed();
            ps = connection.prepareStatement(update);
            setQueryParameters(ps, parameters);
            ps.executeUpdate();
        } catch (SQLException ex) {
            rollbackIfTransactionIsActive();
            throw new SQLMapperException("Error running update (" + update + ")", ex);
        } finally {
            close(rs);
            close(ps);
        }
    }

    /**
     * Begins a new database transaction.
     *
     * @param level transaction isolation level
     * @throws SQLMapperException if a database access error occurs
     */
    public void beginTransaction(TransactionLevel level) throws SQLMapperException {
        try {
            failIfConnectionIsClosed();
            failIfTransactionIsActive();
            connection.setTransactionIsolation(level.getLevel());
            connection.setAutoCommit(false);
            transactionActive = true;
            transactionFailed = false;
        } catch (SQLException ex) {
            transactionFailed = true;
            throw new SQLMapperException("Error beginning transaction", ex);
        }
    }

    /**
     * Ends a database transaction.
     *
     * @throws SQLMapperException if a database access error occurs
     */
    public void endTransaction() throws SQLMapperException {
        try {
            failIfTransactionIsNotActive();
            transactionActive = false;
            if (!transactionFailed) {
                connection.commit();
            }
            connection.setAutoCommit(true);
        } catch (SQLException ex) {
            throw new SQLMapperException("Error ending transaction", ex);
        }
    }

    /**
     * Opens a connection to database or gets a connection from the pool if a
     * pooled data source was passed to constructor.
     *
     * @throws SQLMapperException if a database access error occurs
     */
    public void openConnection() throws SQLMapperException {
        try {
            failIfConnectionIsOpened();
            connection = dataSource.getConnection();
        } catch (SQLException ex) {
            throw new SQLMapperException("Error openning connection to database", ex);
        }
    }

    /**
     * Closes a connection to database or returns a connection to the pool if a
     * pooled data source was passed to constructor.
     *
     * @throws SQLMapperException
     */
    public void closeConnection() throws SQLMapperException {
        try {
            failIfConnectionIsClosed();
            connection.close();
        } catch (SQLException ex) {
            throw new SQLMapperException("Error closing connection to database", ex);
        }
    }

    private <E> List<E> select(String query, Class<E> resultClass, boolean singleResult, Object... parameters) throws SQLMapperException {
        List<E> results = new ArrayList<>();
        PreparedStatement ps = null;
        ResultSet rs = null;
        try {
            failIfConnectionIsClosed();
            ps = connection.prepareStatement(query);
            setQueryParameters(ps, parameters);
            rs = ps.executeQuery();
            if (rs.getMetaData().getColumnCount() > 1) {
                throw new QueryResultException("Query returned more than one column");
            }
            while (rs.next()) {
                if (singleResult && !results.isEmpty()) {
                    throw new QueryResultException("Query returned more than one result");
                }
                results.add(resultClass.cast(rs.getObject(1)));
            }
        } catch (SQLException | QueryResultException ex) {
            rollbackIfTransactionIsActive();
            throw new SQLMapperException("Error running query (" + query + ")", ex);
        } finally {
            close(rs);
            close(ps);
        }
        return results;
    }

    private <E> Object select(String query, ResultMapper<E> mapper, Object... parameters) throws SQLMapperException {
        PreparedStatement ps = null;
        ResultSet rs = null;
        try {
            failIfConnectionIsClosed();
            ps = connection.prepareStatement(query);
            setQueryParameters(ps, parameters);
            rs = ps.executeQuery();
            mapper.setResultSet(rs);
            return mapper.getQueryResult();
        } catch (SQLException | QueryResultException ex) {
            rollbackIfTransactionIsActive();
            throw new SQLMapperException("Error running query (" + query + ")", ex);
        } finally {
            close(rs);
            close(ps);
        }
    }

    private void failIfTransactionIsActive() throws SQLException {
        if (transactionActive) {
            throw new SQLException("Transaction is already active");
        }
    }

    private void failIfTransactionIsNotActive() throws SQLException {
        if (!transactionActive) {
            throw new SQLException("Transaction is not active");
        }
    }

    private void rollbackIfTransactionIsActive() {
        if (transactionActive) {
            transactionFailed = true;
            try {
                connection.rollback();
            } catch (SQLException ex) {
                throw new SQLMapperException("Error rollbacking transaction", ex);
            }
        }
    }

    private void failIfConnectionIsClosed() throws SQLException {
        if (connection == null || connection.isClosed()) {
            throw new SQLException("Connection is closed");
        }
    }

    private void failIfConnectionIsOpened() throws SQLException {
        if (connection != null && !connection.isClosed()) {
            throw new SQLException("Connection is already opened");
        }
    }

    private void setQueryParameters(PreparedStatement ps, Object... parameters) throws SQLException {
        for (int i = 0; i < parameters.length; i++) {
            ps.setObject(i + 1, parameters[i]);
        }
    }

    private void close(Statement s) throws SQLMapperException {
        try {
            if (s != null && !s.isClosed()) {
                s.close();
            }
        } catch (SQLException ex) {
            throw new SQLMapperException("Error closing statement", ex);
        }
    }

    private void close(ResultSet rs) throws SQLMapperException {
        try {
            if (rs != null && !rs.isClosed()) {
                rs.close();
            }
        } catch (SQLException ex) {
            throw new SQLMapperException("Error closing result set", ex);
        }
    }
}
