/*
 * Copyright 2013 Mariusz Kopylec
 *
 * This file is part of SQL Mapper.
 *
 * SQL Mapper is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * SQL Mapper is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Foobar. If not, see http://www.gnu.org/licenses/
 */
package net.sourceforge.kopylec.sqlmapper.core;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import net.sourceforge.kopylec.sqlmapper.exception.QueryResultException;

/**
 * A basic row mapper which maps results of an SQL query to a single object
 * (entity). This class should be extended in order to process a result of an
 * SQL query which returns <b>one or none entity</b>.
 *
 * @author Mariusz Kopylec
 * @param <T> a type to map an SQL query result to
 * @version 1.0
 */
public abstract class SingleResultMapper<T> extends ResultMapper<T> {

    private List<T> results;

    @Override
    T getQueryResult() throws SQLException, QueryResultException {
        results = new ArrayList<>(1);
        while (getResultSet().next()) {
            mapRow();
        }
        if (results.isEmpty()) {
            return null;
        } else if (results.size() > 1) {
            throw new QueryResultException("Query returned more than one entity");
        } else {
            return results.get(0);
        }
    }

    /**
     * Adds an entity to results list if it is not already added. For the method
     * to work correctly an entity must override {@code equals} and
     * {@code hashCode} methods.
     *
     * @param entity entity to merge with results list
     * @return an entity added to results list
     */
    protected final T mergeWithResults(T entity) {
        int entityIndex = results.indexOf(entity);
        if (entityIndex > -1) {
            return results.get(entityIndex);
        } else {
            results.add(entity);
            return entity;
        }
    }

    /**
     * Adds an entity to results list.
     *
     * @param entity entity to add
     */
    protected final void addToResults(T entity) {
        results.add(entity);
    }

    /**
     * Gets the current results list. The results list contains already mapped
     * SQL query results.
     *
     * @return results list
     */
    protected final List<T> getResults() {
        return results;
    }

    /**
     * Maps a current row from result set to the results list.
     *
     * @throws SQLException if a database access error occurs
     */
    protected abstract void mapRow() throws SQLException;
}
