/*
 * Copyright 2013 Mariusz Kopylec
 *
 * This file is part of SQL Mapper.
 *
 * SQL Mapper is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * SQL Mapper is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Foobar. If not, see http://www.gnu.org/licenses/
 */
package net.sourceforge.kopylec.sqlmapper.dao;

import javax.sql.DataSource;
import net.sourceforge.kopylec.sqlmapper.core.SQLRunner;
import net.sourceforge.kopylec.sqlmapper.core.TransactionLevel;
import net.sourceforge.kopylec.sqlmapper.exception.SQLMapperException;

/**
 * A convenient abstract class which wraps an {@link SQLRunner} into a data
 * access object. {@code DataAccessObject} should be extended by more specific
 * user DAOs.
 *
 * @author Mariusz Kopylec
 * @version 1.0
 */
public abstract class DataAccessObject {

    private final SQLRunner sqlRunner;

    public DataAccessObject(DataSource dataSource) {
        sqlRunner = new SQLRunner(dataSource);
    }

    /**
     * Gets an associated instance of {@code SQLRunner}.
     *
     * @return SQL runner
     */
    protected SQLRunner getSqlRunner() {
        return sqlRunner;
    }

    /**
     * Wraps {@link SQLRunner#openConnection()}.
     *
     * @throws SQLMapperException
     */
    public void connect() throws SQLMapperException {
        sqlRunner.openConnection();
    }

    /**
     * Wraps {@link SQLRunner#closeConnection()}.
     *
     * @throws SQLMapperException
     */
    public void disconnect() throws SQLMapperException {
        sqlRunner.closeConnection();
    }

    /**
     * Wraps
     * {@link SQLRunner#beginTransaction(net.sourceforge.kopylec.sqlmapper.core.TransactionLevel)}
     * with a transaction isolation level set to
     * {@link TransactionLevel#READ_COMMITTED}.
     *
     * @throws SQLMapperException
     */
    public void beginTransaction() throws SQLMapperException {
        beginTransaction(TransactionLevel.READ_COMMITTED);
    }

    /**
     * Wraps
     * {@link SQLRunner#beginTransaction(net.sourceforge.kopylec.sqlmapper.core.TransactionLevel)}
     *
     * @param level transaction isolation level
     * @throws SQLMapperException
     */
    public void beginTransaction(TransactionLevel level) throws SQLMapperException {
        sqlRunner.beginTransaction(level);
    }

    /**
     * Wraps {@link SQLRunner#endTransaction()}
     *
     * @throws SQLMapperException
     */
    public void endTransaction() throws SQLMapperException {
        sqlRunner.endTransaction();
    }

    /**
     * Connects to database and opens a transaction with a transaction isolation
     * level set to {@link TransactionLevel#READ_COMMITTED}.
     *
     * @throws SQLMapperException
     */
    public void openSession() throws SQLMapperException {
        openSession(TransactionLevel.READ_COMMITTED);
    }

    /**
     * Connects to database and opens a transaction.
     *
     * @param level transaction isolation level
     * @throws SQLMapperException
     */
    public void openSession(TransactionLevel level) throws SQLMapperException {
        connect();
        beginTransaction(level);
    }

    /**
     * Ends a transaction and disconnects from database.
     *
     * @throws SQLMapperException
     */
    public void closeSession() throws SQLMapperException {
        endTransaction();
        disconnect();
    }
}
